# -*- Mode: Python -*-
# -*- coding: ascii -*-

''' This is a lightwave generic plugin for layout used
to adjust the vertical position (y value) of selected objects
to rest on the chosen triangulated ground object.
I use it whenever the ground surface in my model is irregular, 
like the real world, and I have lots of trees or cars to position. 
It also can be used to set the y values of a camera or other 
object to skim over the irregular surface at a set offset distance

Copyright (c) 2012 John Gutwin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.'''

from __future__ import division

import lwsdk
import operator, random, math, datetime

__author__     = "<John Gutwin>"
__copyright__  = "Copyright (C) 2012 <John Gutwin>"
__version__    = "3.2"
__maintainer__ = "<John Gutwin>"
__email__      = "<jgutwin@pepperchrome.com>"
__status__     = "Production"
__lwver__      = "11"


class CalcPoint(object):
    def __init__(self, x=0.0, y=0.0, z=0.0):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
    def __sub__(self, other):
        return CalcPoint(self.x - other.x, self.y - other.y, self.z - other.z)
    def dot_2(self, other):
        return self.x*other.x + self.z*other.z
class Item():
    def __init__(self, sdk_Info, ID, pCount, ctime = 0):
        self.ID = ID
        self._RIGHT = sdk_Info.param(ID, lwsdk.LWIP_RIGHT, ctime)
        self._UP = sdk_Info.param(ID, lwsdk.LWIP_UP, ctime)
        self._FORWARD = sdk_Info.param(ID, lwsdk.LWIP_FORWARD, ctime)
        self._W_POS = sdk_Info.param(ID, lwsdk.LWIP_W_POSITION, ctime)
        self._PIVOT = sdk_Info.param(ID, lwsdk.LWIP_PIVOT, ctime)
        self._pCount = pCount
        self.P =  None
class Point():
    def __init__(self, pnt, position = None):
        self.i = 0
        self.ID = str(pnt)
        self.pnt = pnt
        self.position = position
class Polygon():
    def __init__(self, polygon_obj, pointCount = None, normal = None, surface = None):
        self.ID = str(polygon_obj)
        self.polygon_obj = polygon_obj
        self.pointCount = pointCount
        self.surface = surface
        self.points = [] #list of CalcPoints
        self.norm = normal
    def prepisItemIn(self):
        #for quicktest - bounding box
        xList = [self.A.x, self.B.x, self.C.x]
        xList.sort()
        self.minX = xList[0]
        self.maxX = xList[2]
        zList = [self.A.z, self.B.z, self.C.z]
        zList.sort()
        self.minZ = zList[0]
        self.maxZ = zList[2]
        
        #for final test
        self.v0 = self.C - self.A
        self.v1 = self.B - self.A
        
        self.dot00 = self.v0.dot_2(self.v0)
        self.dot01 = self.v0.dot_2(self.v1)
        
        self.dot11 = self.v1.dot_2(self.v1)
        self.denom = (self.dot00 * self.dot11 - self.dot01 * self.dot01)
    def isItemIn(self, P):
        # Determine if a point is inside a given traiangle or not (assumes 3 points)
        # code adapted from:  http://www.blackpawn.com/texts/pointinpoly/default.html
        
        if self.denom == 0:
            return False, 0
        if self.minX<P.x<self.maxX and self.minZ<P.z<self.maxZ:#quicktest
            #unchanging terms are pre-calculated in prepisItemIn()
            ## Compute vectors
            #self.v0 = self.C - self.A
            #self.v1 = self.B - self.A
            self.v2 = P - self.A
            ## Compute dot products
            #self.dot00 = self.v0.dot_2(self.v0)
            #self.dot01 = self.v0.dot_2(self.v1)
            self.dot02 = self.v0.dot_2(self.v2)
            #self.dot11 = self.v1.dot_2(self.v1)
            self.dot12 = self.v1.dot_2(self.v2)
            ## Compute barycentric coordinates
            #self.denom = (self.dot00 * self.dot11 - self.dot01 * self.dot01)
            invDenom = 1 / self.denom
            u = (self.dot11 * self.dot02 - self.dot01 * self.dot12) * invDenom
            v = (self.dot00 * self.dot12 - self.dot01 * self.dot02) * invDenom
            if (u >= 0) and (v >= 0) and (u + v < 1): #inside triangle
                ## interpolate elevation
                return True, (u*self.C.y+v*self.B.y+(1.0-(u+v))*self.A.y)
        return False, None
class ground3py(lwsdk.IGeneric):
    def __init__(self, context):
        super(ground3py, self).__init__()
        self._surfaceObject = ''
        self._rotateH = False
        self._rotatePB = False
        self._inclChildren = False
        self._startFrame = 0
        self._endFrame = 0
        self._frameIncrement = 1
        self._offset = 0.0
        self._inclNulls = False
        self.ctime = 0
        self.framesPerSecond = 30
    def process(self, ga):
        ui = lwsdk.LWPanels()
        panel = ui.create('Move selected objects to ground surface - GTO v3.2 python script')

        c0 = panel.item_ctl("Pick a triangulated ground surface", lwsdk.LWI_OBJECT)
        c1 = panel.bool_ctl('Add random heading (good for plants)')
        c2 = panel.bool_ctl('Rotate pivot pitch and bank to match surface (good for parking lots)')
        c3 = panel.bool_ctl('Include hierarchy below selected items')
        c7 = panel.float_ctl('surface offset  - in meters')
        c8 = panel.bool_ctl('Include nulls')
        c9 = panel.text_ctl('the following: ',['"moving items" are incompatable with Rotate pitch and bank'])
        c4 = panel.int_ctl('moving items - Start frame')
        c5 = panel.int_ctl('moving items -   End frame')
        c6 = panel.int_ctl('moving items -          Step')

        panel.align_controls_vertical([c0, c1, c2, c3, c7, c8, c9, c4, c5, c6])
        
        def rotateCallback(controlID, user_data = None):
            #if user choses rotate pivot then disable moving item options
            rotate = controlID.get_int()
            if rotate:
                c4.ghost()
                c5.ghost()
                c6.ghost()
            else:
                c4.render()
                c5.render()
                c6.render()
        c1.set_int(self._rotateH)
        c2.set_int(self._rotatePB)
        c3.set_int(self._inclChildren)
        c4.set_int(self._startFrame)
        c5.set_int(self._endFrame)
        c6.set_int(self._frameIncrement)
        c7.set_float(self._offset)
        c8.set_int(self._inclNulls)
        c2.set_event(rotateCallback)
 
        if panel.open(lwsdk.PANF_BLOCKING | lwsdk.PANF_CANCEL | lwsdk.PANF_MOUSETRAP) == 0:
            ui.destroy(panel)
            return lwsdk.AFUNC_OK

        self._surfaceObject  = c0.get_addr()
        self._rotateH= c1.get_int()
        self._rotatePB= c2.get_int()
        self._inclChildren = c3.get_int()
        if not self._rotatePB:#if user choses rotate pivot then use default moving item options
            self._startFrame = c4.get_int()
            self._endFrame = c5.get_int()
            self._frameIncrement = c6.get_int()
        self._offset = c7.get_float()
        self._inclNulls = c8.get_int()
        
        if self._surfaceObject == None:
            lwsdk.LWMessageFuncs().error("You must choose a surface mesh for this to work ", '')
            return lwsdk.AFUNC_OK # quit without doing anything
            
        self.i_info = lwsdk.LWItemInfo()
        self.o_info = lwsdk.LWObjectInfo()
        self.s_info = lwsdk.LWSceneInfo()
        self.framesPerSecond = self.s_info.framesPerSecond

        numTriangles = self.o_info.numPolygons(self._surfaceObject)
        self.frameCount = 1 + self._endFrame - self._startFrame
        if self._frameIncrement == 0: 
            self._frameIncrement = 1 # 0 would cause problems
        if self._endFrame < self._startFrame:
            self._endFrame = self._startFrame
            lwsdk.LWMessageFuncs().error("forcing end frame equal to start frame ", self._startFrame)
        self.frameCount = 1 + (self._endFrame - self._startFrame)/self._frameIncrement
        
        startTime = datetime.datetime.now()
        #check autokey must be on
        autokeyFlag = lwsdk.LWInterfaceInfo().autoKeyCreate
        if autokeyFlag == lwsdk.LWAKC_OFF: result = ga.evaluate(ga.data,"AutoKey")
        
        self.monf = lwsdk.LWLMonFuncs()
        self.mon = self.monf.create()

        # From selected items make a list of unique top level parents by dropping children of selected items
        selectedItems = lwsdk.LWInterfaceInfo().selected_items()
        topLevelParents = []
        finalItems = []
        if self._inclChildren:
            for item in selectedItems:
                parent= self.i_info.parent(item)
                item_has_no_selected_parent = True
                while parent:
                    if parent in selectedItems:
                        item_has_no_selected_parent = False
                        break
                    parent = self.i_info.parent(parent )
                if item_has_no_selected_parent:
                    topLevelParents.append(item)
            # From unique top level parents make a final ordered list of rest of hierarchy
            for item in topLevelParents: # exchange each item ID for an Item object with additional data
                self.appendItemThenChildren(item, finalItems)
        else:
            for item in selectedItems:
                pCount = self.o_info.numPoints(item)
                if pCount > 2:finalItems.append(Item(self.i_info, item, pCount))
        # Monitor Setup
        if self.mon: 
            text = "Grounding %d objects on %d triangles for %d frames" % \
                (len(finalItems), numTriangles,self.frameCount)
            total = int(len(finalItems)*numTriangles*self.frameCount)
            self.monf.setup( self.mon, text, 0 , None )
            self.monf.init( self.mon, total, 'preparing data...' )
        
        for item in finalItems:
            lwsdk.command("RemoveFromSelection %d" % (item.ID))
        firstFrame = True
        # The surface mesh is being evaluated only at time 0 - move this section to after 'for frame' if mesh is deforming with time
        pCount = self.o_info.numPoints(self._surfaceObject)
        self._surfaceObjectItem = Item(self.i_info, self._surfaceObject, pCount, self.ctime)
        surfaceMesh = self.o_info.meshInfo(self._surfaceObjectItem.ID, 0)
        self.polygons = []
        self.points = []
        self.ptsById = {}
        surfaceMesh.scanPolys(self.scanPolys, (self.polygons,))
        surfaceMesh.scanPoints(self.scanPoints, (self.points,))
        for point in self.points:
            self.ptsById[point.ID] = point
            #get transformed (if any) surface points position and translate to world coords
            position = surfaceMesh.pntOtherPos(point.pnt)
            point.position = CalcPoint(*self.convertToWorld(self._surfaceObjectItem , position))
        for poly in self.polygons:
            # create a list of the points in just this polygon (we are only interested in 3)
            poly.pointCount = surfaceMesh.polSize(poly.polygon_obj)
            poly.A = self.ptsById[str(surfaceMesh.polVertex(poly.polygon_obj, 0))].position
            poly.B = self.ptsById[str(surfaceMesh.polVertex(poly.polygon_obj, 1))].position
            poly.C = self.ptsById[str(surfaceMesh.polVertex(poly.polygon_obj, 2))].position
            #fetch normal
            normal = surfaceMesh.polOtherNormal(poly.polygon_obj)[1]
            #store normal as CalcPoint
            poly.norm = CalcPoint(normal.x, normal.y, normal.z)
            #pre-calculate as many terms as possible before looping 
            poly.prepisItemIn()
        for frame in xrange(self._startFrame, self._endFrame+1, self._frameIncrement):
            self.ctime = frame / self.framesPerSecond
            result = ga.evaluate(ga.data,"GoToFrame %d" % (frame))
            self.monf.step( self.mon, 0, "Starting Frame %d" % (frame) )
            for itemCount, item in enumerate(finalItems):
                if not self._inclNulls: # are nulls included?
                    if item._pCount == 0: # NULL items have no points
                        continue # skip NULL item
                curTime = datetime.datetime.now()
                #update world position of item for current time
                item._W_POS = self.i_info.param(item.ID, lwsdk.LWIP_W_POSITION, self.ctime)
                item.P = CalcPoint(item._W_POS[0],0,item._W_POS[2])
                for polyCount, poly in enumerate(self.polygons):
                    if self.monf.step( self.mon, 1, None ):
                        self.monf.step( self.mon, 0, "Aborted!" )
                        self.aborted = True
                        return lwsdk.AFUNC_OK
                    #Test each item - is it in a triangle of the mesh
                    found, Ys = poly.isItemIn(item.P)
                    if found:
                        if self.monf.step( self.mon, numTriangles-polyCount, \
                            'Found item # %d on surface triangle # %d   frame %d' %\
                            (itemCount ,numTriangles-polyCount, frame) ):
                            self.monf.step(  self.mon, 0, "Aborted!" )
                            self.aborted = True
                            return lwsdk.AFUNC_OK
                        # adjust elevation for parent world elevation and user chosen offset
                        parentID = self.i_info.parent(item.ID)
                        Parent_W_POS = self.i_info.param(parentID, lwsdk.LWIP_W_POSITION, self.ctime)
                        adjustedElevation = Ys - Parent_W_POS[1] + self._offset
                        # fetch items envelope and channel ID's
                        envef = lwsdk.LWEnvelopeFuncs()
                        chinfo = lwsdk.LWChannelInfo()
                        groupID = self.i_info.chanGroup( item.ID )
                        chanID = chinfo.nextChannel(groupID, None)
                        while chanID:
                            channelName = chinfo.channelName(chanID)
                            if channelName == "Position.Y":
                                chanY = chanID
                            if channelName == "Rotation.H":
                                chanH = chanID
                            chanID = chinfo.nextChannel(groupID, chanID)
                        #if not chanID...     its possible the channel does not exist
                        key = envef.findKey( chanY, self.ctime )
                        if key:
                            envef.keySet(chanY, key, lwsdk.LWKEY_VALUE, adjustedElevation)
                        else:
                            key = envef.createKey( chanY, self.ctime, adjustedElevation)
                        if firstFrame and self._rotateH: #firstFrame only -  animated random rotation would be silly
                            key = envef.findKey( chanH, self.ctime )
                            if key:
                                envef.keySet(chanY, key, lwsdk.LWKEY_VALUE, random.random()*4)
                            else:
                                key = envef.createKey( chanH, self.ctime, random.random()*4)
                        if self.frameCount==1 and self._rotatePB: # only single frame is rotated since pivot rotation has no envelope
                            p = math.degrees(math.asin(poly.norm.z))
                            b = math.degrees(math.asin(-poly.norm.x))
                            result = ga.evaluate(ga.data,"SelectItem %x" % (item.ID))
                            result = ga.evaluate(ga.data,"PivotRotation %f %f %f" % (0, p,  b))
                        break
            firstFrame = False
        #if we turned autokey on we should turn it back off
        if autokeyFlag == lwsdk.LWAKC_OFF: result = ga.evaluate(ga.data,"AutoKey")
        #total time
        elapsedTime = datetime.datetime.now() - startTime
        elapsedTime = str(elapsedTime)[:-6] #trim microseconds off
        print "GTO Grounded %d objects on %d triangles for %d frames in %s (h:m:s) Time" %\
            (len(finalItems), numTriangles,self.frameCount,elapsedTime)
        if ( self.monf and self.mon ): self.monf.done( self.mon )
        if ( self.monf and self.mon ): self.monf.destroy( self.mon )
        return lwsdk.AFUNC_OK
    def appendItemThenChildren(self, item, theList):
        """ re-entrant code to step through the hierarchy ensuring the
        children follow the parents
        """
        if not item: return
        # exchange each item ID for an Item object with additional data
        pCount = self.o_info.numPoints(item)
        if pCount > 2:theList.append(Item(self.i_info, item, pCount))
        child = self.i_info.firstChild(item )
        self.appendItemThenChildren(child, theList)
        while child:
            nextChild = self.i_info.nextChild(item, child)
            self.appendItemThenChildren(nextChild, theList)
            child = nextChild
    def scanPolys(self, poly_list, poly_id):
        poly_list.append(Polygon(poly_id))
        return lwsdk.EDERR_NONE
    def scanPoints(self, point_list, point_id):
        point_list.append(Point(point_id))
        return lwsdk.EDERR_NONE
    def convertToWorld(self, item, p):
        q = []
        for i in range(3):
            q.append(( p[0] - item._PIVOT[0] ) * item._RIGHT[ i ] + \
                        ( p[1] - item._PIVOT[1] ) * item._UP[ i ] + \
                        ( p[2] - item._PIVOT[2] ) * item._FORWARD[ i ]+ item._W_POS[ i ])
        return q

ServerTagInfo = [
                    ( "Ground those objects 3", lwsdk.SRVTAG_USERNAME | lwsdk.LANGID_USENGLISH ),
                    ( "GTO 3", lwsdk.SRVTAG_BUTTONNAME | lwsdk.LANGID_USENGLISH ),
                    ( "Utilities/Python", lwsdk.SRVTAG_MENU | lwsdk.LANGID_USENGLISH )
                ]

ServerRecord = { lwsdk.GenericFactory("LW_Pyground3py", ground3py) : ServerTagInfo }